//
//  Frame I/O and bitstream functions
//
#include <assert.h>
#include "defines.h"


//
//
//
int read_one_new_frame(int height, int width, int skip, int start, int imgNo, FILE *p_in, imgStruct *img)
{
  int i, j, uv;
  long status = 0; 
  int frameNo;
  int frameSize = height * width * 3 / 2;
  static byte *buf = NULL;
  byte *tmp_buf = NULL;
  static int first = 1;

  if(first)
  {
    buf = (byte *) malloc(frameSize * sizeof(byte));
    first = 0;
  }

  frameNo = imgNo * (skip + 1) + start;
   
  status |= fseek(p_in, frameNo * frameSize, SEEK_SET);
  status |= (fread(buf, frameSize * sizeof(byte), 1, p_in) != 1);

  if(!status)
  {
    tmp_buf = buf;

    for(j=0; j < height; j++)
      for(i=0; i < width; i++)
        img->imgY[j][i] = (unsigned short) *tmp_buf++;

    for(uv=0; uv < 2; uv++)
      for(j=0; j < height/2 ; j++)
        for(i=0; i < width/2; i++)
          img->imgUV[uv][j][i] = (unsigned short) *tmp_buf++;
  }
  else
  {
    printf("Error reading frame\n");
  }

  return (status);
}

//
//
//
void write_image(int height, int width, FILE *p_out, imgStruct img)
{
  int i, j, k;
  int frameSize = height * width * 3 / 2;
  static byte *buf = NULL;
  byte *tmp_buf = NULL;
  static int first = 1;

  if(first)
  {
    buf = (byte *) malloc(frameSize * sizeof(byte));
    first = 0;
  }
    
  tmp_buf = buf;

  for(i=0; i < height; i++)
    for(j=0; j < width; j++)
      *tmp_buf++ = (byte) img.imgY[i][j];
    
  for(k=0; k < 2; ++k)
    for(i=0; i < height/2; i++)
      for(j=0; j < width/2; j++)
      *tmp_buf++ = (byte) img.imgUV[k][i][j];

  fwrite(buf, frameSize * sizeof(byte), 1, p_out);
}

//
//
//
Bitstream * allocBitstream(int size)
{
  Bitstream *tmp;

  tmp = (Bitstream *) calloc(1, sizeof(Bitstream));
  tmp->streamBuffer = (byte *) calloc(size, sizeof(byte));
  tmp->bits_to_go = 8;
  tmp->byte_buf = 0;
  tmp->byte_pos = 0;

  return tmp;
}

//
//
//
int freeBitstream(Bitstream *bitstream)
{
  int tmp = 0;
  tmp += bitstream->byte_pos * 8;
  tmp += 8 - bitstream->bits_to_go;

  free(bitstream->streamBuffer);
  bitstream->streamBuffer = NULL;
  free(bitstream);

  return tmp;
}


//
//  Pad and save bitstream on file
//
void saveBitstream(Bitstream *bitstream, char *fname)
{
  FILE *fp;
  int cnt = 0;

  if(bitstream->bits_to_go != 8 )
    while(bitstream->bits_to_go != 8)
    {
      u_1("zero padding", 0, bitstream); 
      ++cnt;
    }

  if((fp = fopen(fname, "wb")) == 0)
  {
    printf("Error open file %s\n", fname);
    exit(0);
  }  

  if(bitstream->byte_pos != fwrite(bitstream->streamBuffer, sizeof(byte), bitstream->byte_pos, fp))
  {
    printf("Error writing file %s\n", fname);
    exit(0);
  }  

  fclose(fp);
  printf("Total filter bits = %d (%d padding)\n", bitstream->byte_pos * 8, cnt);
}


//
//
//
void  writeUVLC2buffer(SyntaxElement *se, Bitstream *currStream)
{

  int i;
  unsigned int mask = 1 << (se->len - 1);
  assert((se->len - 1) < 32);

  for(i = 0; i < se->len; i++)
  {
    currStream->byte_buf <<= 1;
    if(se->bitpattern & mask)
      currStream->byte_buf |= 1;
    currStream->bits_to_go--;
    mask >>= 1;
    if(currStream->bits_to_go == 0)
    {
      currStream->bits_to_go = 8;
      currStream->streamBuffer[currStream->byte_pos++] = currStream->byte_buf;
      currStream->byte_buf = 0;
    }
  }
}

//
//
//
int symbol2vlc(SyntaxElement *sym)
{
  int info_len = sym->len;

  // Convert info into a bitpattern int
  sym->bitpattern = 0;

  // vlc coding
  while(--info_len >= 0)
  {
    sym->bitpattern <<= 1;
    sym->bitpattern |= (0x01 & (sym->inf >> info_len));
  }
  return 0;
}

//
//
//
int u_1(char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym = &symbol;

  sym->bitpattern = value;
  sym->len = 1;
  sym->value1 = value;
  assert(bitstream->streamBuffer != NULL);
  writeUVLC2buffer(sym, bitstream);
  return (sym->len);
}

//
//
//
int u_v(int n, char *tracestring, int value, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym = &symbol;

  sym->bitpattern = value;
  sym->len = n;
  sym->value1 = value;
  assert (bitstream->streamBuffer != NULL);
  writeUVLC2buffer(sym, bitstream);
  return (sym->len);
}

//
// from http://en.wikipedia.org/wiki/Golomb_coding
// M=2^k
//
int golombEncode(int coeff, int k, Bitstream *bitstream, char *tracestring)
{
  int q, i, m;
  int len = 0;
  int symbol = abs(coeff);

  m = (int)pow(2.0, k);
  q = symbol / m;

  for (i = 0; i < q; i++)
    len += u_1(tracestring, 1, bitstream);    // write q ones
  len += u_1(tracestring, 0, bitstream);      // write one zero

  for(i = 0; i < k; i++)
  {
    len += u_1(tracestring, symbol & 0x01, bitstream); 
    symbol >>= 1;
  }

  if(coeff != 0)
  {
    int sign = (coeff > 0)? 1: 0;
    len += u_1(tracestring, sign, bitstream); 
  }
  return(len);
}

//
//
//
int lengthGolomb(int coeffVal, int k)
{
  int m = 2 << (k - 1);
  int q = coeffVal / m;
  if(coeffVal != 0)
    return(q + 2 + k);
  else
    return(q + 1 + k);
}

//
//
//
int allOnes(int *v, int len)
{
  int i;
  for(i = 0; i < len; ++i)
    if(v[i] == 0)
      return 0;
  return 1;
}


//
//  force0: 
//    0 - normal
//    1 - force0 flag followed by codedVarBins[0...filters_per_group-1]
//
//  predMethod:
//    0 - filter coefficients non predicted
//    1 - filter coefficients predicted
//
int writePredFlags(int force0, int predMethod, int codedVarBins[NO_VAR_BINS], 
                   int filters_per_group, Bitstream *bitstream, int createBitstream)
{
  int bit_cnt = 0;
  int ind;

  if(createBitstream)
  {
    bit_cnt += u_1("force0", force0, bitstream);
    if(force0)
      for(ind = 0; ind < filters_per_group; ++ind)
        bit_cnt += u_1("codedVarBins[i]", codedVarBins[ind], bitstream);
    bit_cnt += u_1("predMethod", predMethod, bitstream);
  }
  else
  {
    if(force0)
      bit_cnt = 2 + filters_per_group;
    else
      bit_cnt = 2;
  }
  return bit_cnt;
}


//
//
//
int writeFilterCodingParams(int minKStart, int maxScanVal, int kMinTab[], Bitstream *bitstream, int createBistream)
{
  int scanPos;
  int golombIndexBit;
  int bit_cnt = 0;
  int kMin;

  if(createBistream)
  {
    // Golomb parameters
    bit_cnt += u_v (3, "expGolomb index", minKStart - 1, bitstream);

    kMin = minKStart; 
    for(scanPos = 0; scanPos < maxScanVal; scanPos++)
    {
      golombIndexBit = (kMinTab[scanPos] != kMin)? 1: 0;

      assert(kMinTab[scanPos] <= kMin + 1);

      bit_cnt += u_1("expGolomb index", golombIndexBit, bitstream);
      kMin = kMinTab[scanPos];
    }    
  }
  else
  {
    bit_cnt += 3;                               // minKStart
    bit_cnt += maxScanVal;                      // golombIndexBit[]
  }

  return bit_cnt;
}


//
//
//
int writeFilterCoeffs(int sqrFiltLength, int filters_per_group, int pDepthInt[], 
                      int **FilterCoeff, int kMinTab[], Bitstream *bitstream, int createBitstream)
{
  int ind, scanPos, i;
  int bit_cnt = 0;

  for(ind = 0; ind < filters_per_group; ++ind)
  {
    for(i = 0; i < sqrFiltLength; i++)
    {	
      scanPos = pDepthInt[i] - 1;
      if (createBitstream)
        bit_cnt += golombEncode(FilterCoeff[ind][i], kMinTab[scanPos], bitstream, "full-pel filter");
      else
        bit_cnt += lengthGolomb(abs(FilterCoeff[ind][i]), kMinTab[scanPos]);
    }
  }

  return bit_cnt;
}

//
// 
//
int codeAuxInfo(int filtNo, int noFilters, int varIndTab[NO_VAR_BINS], int usePrevFiltBest[NO_VAR_BINS], 
                int prevFiltInd, int frNo, Bitstream *bitstream, int createBitstream)
{
  int i, predPattern, filterPattern[NO_VAR_BINS], startSecondFilter=0, bitCt=0, codePrediction;
  SyntaxElement symbol, *sym=&symbol;

  int lentab_pred[3]={1, 2, 2};
  int codtab_pred[3]={1, 1, 0};
  int lentab_noFilters[3]={1, 2, 2};
  int codtab_noFilters[3]={1, 1, 0};
  int lentab_filtNo[3]={2, 2, 1};
  int codtab_filtNo[3]={0, 1, 1};

  codePrediction = 0;

  if(frNo > 0)
  {
    codePrediction = 1;

    if(createBitstream == 1)
    {
      if(filtNo >= 0)
        bitCt += u_1("Nonzero coefficients", 1, bitstream);
      else
        bitCt += u_1("Nonzero coefficients", 0, bitstream);
    }
    else
    {
      bitCt += 1;
    }
  }

  // Filter order, number of filters & variance for which they are used
  if(filtNo >= 0)
  {
    sym->len=lentab_filtNo[filtNo];
    if (createBitstream==1){
      sym->inf=codtab_filtNo[filtNo];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt+=sym->len;

    if(noFilters > 1)
    {
      memset(filterPattern, 0, NO_VAR_BINS * sizeof(int)); 
      for(i = 1; i < NO_VAR_BINS; ++i)
        if(varIndTab[i] != varIndTab[i-1])
        {
          filterPattern[i] = 1;
          startSecondFilter = i;
        }
    }

    sym->len=lentab_noFilters[min(noFilters-1, 2)];
    if(createBitstream == 1)
    {
      sym->inf=codtab_noFilters[min(noFilters-1, 2)];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt+=sym->len;

    if(noFilters == 2)
    {
      if(createBitstream == 1)
        u_v (4, "startSecondFilter", startSecondFilter, bitstream);
      bitCt+=4;
    }
    else if(noFilters > 2)
    {
      if(createBitstream == 1)
        for(i = 1; i < NO_VAR_BINS; i++)
          u_1("filterPattern[i]", filterPattern[i], bitstream);
      bitCt+=15;  // Was 16
    }
  }

  if(codePrediction == 1)
  {
    if (createBitstream == 1)
      u_v (4, "prevFiltInd", prevFiltInd, bitstream);
    bitCt+=4;

    // prediction pattern
    predPattern = (usePrevFiltBest[0] == 1)? 0: 1;
    for(i = 1; i < NO_VAR_BINS; i++)
    {
      if(usePrevFiltBest[i] == 0)
      {
        predPattern=2;
        break;
      }
    }

    sym->len = lentab_pred[predPattern];
    if(createBitstream == 1)
    {
      sym->inf = codtab_pred[predPattern];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt += sym->len;

    if(predPattern == 2)
    {
      if(createBitstream == 1)
        for(i = 0; i < NO_VAR_BINS; i++)
          u_1("usePrevFiltBest[i]", usePrevFiltBest[i], bitstream);
      bitCt += 16;
    }
  }
  return(bitCt);
}




